from collections.abc import Mapping
from typing import Generic, Hashable, Iterator, Iterable, TypeVar, get_args
from sowing import traversal
from sowing.node import Node
from sowing.zipper import Zipper
from .util.rangequery import RangeQuery
from dataclasses import field, Field
import inspect


NodeData = TypeVar("NodeData", bound=Hashable)
EdgeData = TypeVar("EdgeData", bound=Hashable)


TreeElement = Zipper[NodeData, EdgeData] | Node[NodeData, EdgeData] | str
TreeKey = str


def get_key(element: TreeElement) -> TreeKey:
    if isinstance(element, Zipper):
        data = element.node.data
    elif isinstance(element, Node):
        data = element.data
    else:
        data = element

    if isinstance(data, Mapping):
        return data.get("name", "")

    if hasattr(data, "name"):
        return getattr(data, "name")

    if isinstance(data, TreeKey):
        return data

    return ""


class IndexedTree(Generic[NodeData, EdgeData]):
    """Structure for fast querying of tree nodes by key."""

    __slots__ = [
        "root",
        "_depths",
        "_to_index",
        "_to_cursor",
        "_all_keys",
        "_all_cursors",
    ]

    def __init__(self, root: Node[NodeData, EdgeData]):
        """
        Initialize an indexed tree.

        Complexity: O(n log n), with n the number of nodes below :param:`root`.

        :param root: root of the input tree to index on
        :raises: if any two nodes share the same key
        """
        to_cursor: dict[str, Zipper[NodeData, EdgeData]] = {}
        all_keys: list[str] = []
        all_cursors: list[Zipper[NodeData, EdgeData]] = []

        for cursor in traversal.depth(root, preorder=True):
            key = get_key(cursor)

            if key in to_cursor:
                raise RuntimeError(f"duplicate key {key!r} in tree {root!r}")

            to_cursor[cursor] = cursor
            to_cursor[key] = cursor
            all_cursors.append(cursor)
            all_keys.append(key)

        to_index: dict[Zipper[NodeData, EdgeData], int] = {}
        depths: list[tuple[int, Zipper[NodeData, EdgeData]]] = []

        for cursor in traversal.euler(root):
            to_index[cursor] = len(depths)
            to_index[get_key(cursor)] = len(depths)
            depths.append((cursor.depth, cursor))

        self.root = root
        self._to_cursor = to_cursor
        self._to_index = to_index
        self._depths = RangeQuery(depths, min)
        self._all_keys = all_keys
        self._all_cursors = all_cursors

    def __call__(self, *keys: TreeElement) -> Zipper[NodeData, EdgeData]:
        """
        Locate a node by its key or the lowest common ancestor of a collection of keys.

        Complexity: O(n), the number of arguments.

        :param keys: node key or collection of keys
        :raises TypeError: if no arguments are passed
        :returns: if a single key is passed, return the corresponding node;
            otherwise, return the lowest common ancestor of the collection
            of nodes
        """
        if not keys:
            raise TypeError("at least one node is needed")

        start = end = self._to_index[keys[0]]

        for key in keys[1:]:
            index = self._to_index[key]
            start = min(start, index)
            end = max(end, index)

        result = self._depths(start, end + 1)
        assert result is not None
        return result[1]

    def __getitem__(self, key: TreeElement) -> Zipper[NodeData, EdgeData]:
        """Locate a tree position by its key."""
        return self._to_cursor[key]

    def __contains__(self, key: TreeElement) -> bool:
        return key in self._to_cursor

    def __len__(self) -> int:
        """Get the number of nodes in the tree."""
        return len(self._all_keys)

    def __iter__(self) -> Iterator[NodeData]:
        """Iterate through the keys of all nodes in the tree."""
        return iter(self._all_keys)

    def keys(self) -> Iterable[str]:
        return self._all_keys

    def values(self) -> Iterable[Zipper[NodeData, EdgeData]]:
        return self._all_cursors

    def items(self) -> Iterable[tuple[str, Zipper[NodeData, EdgeData]]]:
        return zip(self._all_keys, self._all_cursors)

    def is_ancestor_of(self, key1: TreeElement, key2: TreeElement) -> bool:
        """
        Check whether a node is an ancestor of another.

        Complexity: O(1).

        :returns: True if and only if :param:`key2` is on the path from the tree
            root to :param:`key1`
        """
        return self(key1, key2) == self(key1)

    def is_strict_ancestor_of(self, key1: TreeElement, key2: TreeElement) -> bool:
        """
        Check whether a node is a strict an ancestor of another
        (i.e. is an ancestor distinct from the other node).

        Complexity: O(1).

        :returns: True if and only if :param:`key2` is on the path from the tree
            root to :param:`key1` and different from :param:`key1`
        """
        return self(key1, key2) == self(key1) and key1 != key2

    def is_comparable(self, key1: TreeElement, key2: TreeElement) -> bool:
        """
        Check whether two nodes are in the same subtree.

        Complexity: O(1).

        :returns: True if and only if either :param:`key1` is an ancestor or
            descendant of :param:`key2`
        """
        return self.is_ancestor_of(key1, key2) or self.is_ancestor_of(key2, key1)

    def depth(self, key: TreeElement) -> int:
        """
        Find the depth of a node.

        Complexity: O(1).
        """
        index = self._to_index[self[key]]
        return self._depths(index, index + 1)[0]

    def distance(self, key1: TreeElement, key2: TreeElement) -> int:
        """
        Compute the number of edges on the shortest path between two nodes.

        Complexity: O(1).
        """
        return self.depth(key1) + self.depth(key2) - 2 * self.depth(self(key1, key2))


def index_trees(cls):
    """
    Create index caches for selected tree fields in a dataclass.

    This is a decorator that should be invoked before dataclass(),
    i.e., placed after @dataclass() in the decorator list.

    All fields having the "index_tree" metadata key will be turned into index
    caches for the specified tree fields.
    """
    mapping = {}

    for name, decl in inspect.getmembers(cls):
        if isinstance(decl, Field) and "index_from_tree" in decl.metadata:
            target_field = decl.metadata["index_from_tree"]
            mapping[name] = target_field
            setattr(cls, name, field(init=False, repr=False, compare=False))

            target_type = cls.__annotations__[target_field]
            cls.__annotations__[name] = IndexedTree[*get_args(target_type)]

    if mapping:
        orig_postinit = getattr(cls, "__post_init__", lambda self: None)

        def cls_postinit(self):
            orig_postinit(self)

            for indexed, orig in mapping.items():
                object.__setattr__(self, indexed, IndexedTree(getattr(self, orig)))

        cls.__post_init__ = cls_postinit

    return cls



from dataclasses import fields


def repr_default(cls):
    """Replace a dataclass __repr__ to omit fields with default values."""

    def cls_repr(self):
        args = ", ".join(
            f"{field.name}={getattr(self, field.name)!r}"
            for field in fields(cls)
            if getattr(self, field.name) != field.default and field.repr
        )
        return f"{cls.__qualname__}({args})"

    cls.__repr__ = cls_repr
    return cls



from typing import Any, Iterator
from collections import deque
from enum import Enum, auto
from dataclasses import dataclass
from immutables import Map
from sowing.node import Node


class ParseError(Exception):
    """Parsing error with location information."""

    def __init__(self, message, start, end):
        if abs(start - end) <= 1:
            message = f"{message} (at position {start})"
        else:
            message = f"{message} (at range {start}-{end})"

        super().__init__(message)
        self.start = start
        self.end = end


class TokenKind(Enum):
    OpenParen = "("
    CloseParen = ")"

    OpenPropsNHX = "[&&NHX"
    OpenPropsBEAST = "[&"
    CloseProps = "]"

    Comma = ","
    Colon = ":"
    Semicolon = ";"
    Equals = "="

    String = "string"
    End = "end"


class PropStyle(Enum):
    """Lexer state for different existing notations for properties."""

    # Outside of any property block
    Normal = auto()

    # Properties using the New Hampshire eXtended (NHX) notation
    # <https://home.cc.umanitoba.ca/~psgendb/doc/atv/NHX.pdf>
    NHX = auto()

    # Properties using the BEAST notation
    # <https://beast.community/nexus_metacomments.html>
    BEAST = auto()


@dataclass(frozen=True)
class Token:
    """Token emitted by the lexer."""

    # Type of token
    kind: TokenKind

    # Token starting position in the data stream
    start: int

    # Position following the token end in the data stream
    end: int

    # Value associated with the token
    value: Any = None


# Class of characters that ignored by the lexer
WHITESPACE = " \t\n"


def _lex_whitespace(data: str, pos: int) -> int:
    """Advance past whitespace in a Newick string."""
    while pos < len(data) and data[pos] in WHITESPACE:
        pos += 1

    return pos


def _lex_comment(data: str, pos: int) -> int:
    """Advance past a comment in a Newick string."""
    start = pos
    pos += 1
    depth = 1
    contents = "["

    while pos < len(data) and depth > 0:
        if data[pos] == "]":
            depth -= 1
        elif data[pos] == "[":
            depth += 1

        contents += data[pos]
        pos += 1

    if depth > 0:
        raise ParseError("unclosed comment", start, pos)

    return pos


def _lex_quoted_string(data: str, pos: int) -> tuple[Token, int]:
    """Extract a token for a string surrounded by single quotes."""
    start = pos
    pos += 1
    contents = ""

    while pos < len(data):
        if data[pos] == "'":
            if pos + 1 < len(data) and data[pos + 1] == "'":
                contents += "'"
                pos += 2
            else:
                break

        contents += data[pos]
        pos += 1

    if pos == len(data):
        raise ParseError("unclosed string", start, pos)

    end = pos + 1
    return Token(TokenKind.String, start, end, contents), end


def _lex_unquoted_string(data: str, state: PropStyle, pos: int) -> tuple[Token, int]:
    """Extract a token for a plain unquoted string."""
    start = pos
    contents = data[pos]
    pos += 1

    match state:
        case PropStyle.Normal:
            break_chars = "()[],:;' \t\n"

        case PropStyle.NHX:
            break_chars = "[]:= \t\n"

        case PropStyle.BEAST:
            break_chars = "[],= \t\n"

    while pos < len(data):
        cur = data[pos]

        if cur in break_chars:
            break

        if cur == "_":
            contents += " "
        else:
            contents += cur

        pos += 1

    return Token(TokenKind.String, start, pos, contents), pos


def tokenize(data: str) -> Iterator[Token]:
    """Tokenize a Newick string."""
    pos = 0
    state = PropStyle.Normal
    nhx_start = TokenKind.OpenPropsNHX
    beast_start = TokenKind.OpenPropsBEAST

    while (pos := _lex_whitespace(data, pos)) < len(data):
        cur = data[pos]

        match state:
            case PropStyle.Normal:
                token_chars = "(),:;"

            case PropStyle.NHX:
                token_chars = ":="

            case PropStyle.BEAST:
                token_chars = ",="

        if cur in token_chars:
            yield Token(TokenKind(cur), pos, pos + 1)
            pos += 1

        elif data[pos:].startswith(nhx_start.value):
            state = PropStyle.NHX
            yield Token(nhx_start, pos, pos + len(nhx_start.value))
            pos += len(nhx_start.value)

        elif data[pos:].startswith(beast_start.value):
            state = PropStyle.BEAST
            yield Token(beast_start, pos, pos + len(beast_start.value))
            pos += len(beast_start.value)

        elif cur == "]":
            if state == PropStyle.Normal:
                raise ParseError("unexpected ']'", pos, pos + 1)
            else:
                yield Token(TokenKind.CloseProps, pos, pos + 1)
                state = PropStyle.Normal
                pos += 1

        elif cur == "[":
            pos = _lex_comment(data, pos)

        elif cur == "'":
            token, pos = _lex_quoted_string(data, pos)
            yield token

        else:
            token, pos = _lex_unquoted_string(data, state, pos)
            yield token

    yield Token(TokenKind.End, pos, pos)


class TokenIterator:
    """Token iterator with pushback."""

    def __init__(self, iterator: Iterator):
        self.iterator = iterator
        self.buffer = deque()

    def __next__(self):
        if self.buffer:
            return self.buffer.popleft()

        return self.iterator.__next__()

    def push(self, token: Token) -> None:
        """Push a token back into the stream."""
        self.buffer.append(token)

    def skip(self, kind: TokenKind) -> None:
        """Skip over the next token if it is of given type."""
        token = next(self)

        if token.kind != kind:
            self.push(token)

    def extract(self, kind: TokenKind) -> Token | None:
        """Extract the next token if it is of given type, or push it back otherwise."""
        token = next(self)

        if token.kind == kind:
            return token

        self.push(token)

    def expect(self, kind: TokenKind) -> Token:
        """Ensure the next token is of given type and return it."""
        token = next(self)

        if token.kind != kind:
            raise ParseError(
                f"expected '{kind.value}', not '{token.kind.value}'",
                token.start,
                token.end,
            )

        return token


class ParseState(Enum):
    """States for the Newick parser."""

    # At the beginning of a node, ready to read its children
    NodeStart = auto()

    # After a node’s children, ready to read its attached data
    NodeData = auto()

    # At the end of a tree, before the final semicolon
    Finish = auto()


def _parse_props_nhx(tokens: TokenIterator) -> Map:
    """Parse a block of Newick properties in NHX format."""
    result = {}

    while tokens.extract(TokenKind.Colon) is not None:
        key = tokens.expect(TokenKind.String).value
        tokens.expect(TokenKind.Equals)

        if (token := tokens.extract(TokenKind.String)) is not None:
            value = token.value
        else:
            value = ""

        result[key] = value

    tokens.expect(TokenKind.CloseProps)
    return Map(result)


def _parse_props_beast(tokens: TokenIterator) -> Map:
    """Parse a block of Newick properties in BEAST format."""
    result = {}

    while (token := tokens.extract(TokenKind.String)) is not None:
        key = token.value
        tokens.expect(TokenKind.Equals)

        if (token := tokens.extract(TokenKind.String)) is not None:
            value = token.value
        else:
            value = ""

        result[key] = value
        tokens.skip(TokenKind.Comma)

    tokens.expect(TokenKind.CloseProps)
    return Map(result)


def _parse_props(tokens: TokenIterator) -> tuple[Map]:
    """Parse a block of Newick properties."""
    match (start := next(tokens)).kind:
        case TokenKind.OpenPropsNHX:
            return _parse_props_nhx(tokens)

        case TokenKind.OpenPropsBEAST:
            return _parse_props_beast(tokens)

        case _:
            tokens.push(start)
            return Map()


def parse_chain(data: str) -> tuple[Node, int]:
    """
    Chainable parser for single trees encoded as Newick strings.

    :param data: input data stream
    :return: parsed tree and ending position in the string
    """
    nodes = []
    tokens = TokenIterator(tokenize(data))
    state = ParseState.NodeStart

    while state != ParseState.Finish:
        match state:
            case ParseState.NodeStart:
                # Start parsing a new node
                nodes.append(Node())

                if tokens.extract(TokenKind.OpenParen) is not None:
                    state = ParseState.NodeStart
                else:
                    state = ParseState.NodeData

            case ParseState.NodeData:
                # Parse metadata attached to a node
                clade = Map()
                branch = Map()

                # Parse node label
                if (token := tokens.extract(TokenKind.String)) is not None:
                    clade = clade.set("name", token.value)

                # Parse node props
                clade = clade.update(_parse_props(tokens))

                if tokens.extract(TokenKind.Colon) is not None:
                    # Parse branch length
                    if (token := tokens.extract(TokenKind.String)) is not None:
                        branch = branch.set("length", token.value)

                    # Parse branch support
                    if tokens.extract(TokenKind.Colon) is not None:
                        if (token := tokens.extract(TokenKind.String)) is not None:
                            branch = branch.set("support", token.value)

                    # Parse branch probability
                    if tokens.extract(TokenKind.Colon) is not None:
                        if (token := tokens.extract(TokenKind.String)) is not None:
                            branch = branch.set("probability", token.value)

                    # Parse other branch props
                    branch = branch.update(_parse_props(tokens))

                active = nodes.pop()
                active = active.replace(data=clade or None)

                if not nodes:
                    # Finished parsing the root node
                    state = ParseState.Finish
                    nodes.append(active)
                else:
                    # Attach parsed node to its parent
                    parent = nodes.pop()
                    nodes.append(parent.add(active, data=branch or None))

                    match (token := next(tokens)).kind:
                        case TokenKind.Comma:
                            state = ParseState.NodeStart

                        case TokenKind.CloseParen:
                            state = ParseState.NodeData

                        case _:
                            raise ParseError(
                                f"unexpected token '{token.kind.value}' after node",
                                token.start,
                                token.end,
                            )

    token = next(tokens)

    if token.kind != TokenKind.Semicolon:
        raise ParseError(
            f"expected ';' after end of tree, not '{token.kind.value}'",
            token.start,
            token.end,
        )

    return nodes.pop(), token.end


def parse(data: str) -> Node[Map, Map]:
    """Parse a single tree encoded as a Newick string."""
    node, pos = parse_chain(data)

    if data[pos:].strip(WHITESPACE):
        raise ParseError("unexpected garbage after end of tree", pos, len(data))

    return node


def parse_all(data: str) -> list[Node[Map, Map]]:
    """Parse a sequence of trees encoded as Newick strings."""
    result = []

    while data.strip(WHITESPACE):
        node, pos = parse_chain(data)
        result.append(node)
        data = data[pos:]

    return result



def _ilog2(value: int):
    """Integral part of the base-2 logarithm of a positive integer."""
    return value.bit_length() - 1


class RangeQuery:
    """
    Structure for fast computation of indempotent functions on ranges.

    For an input list of N elements, this structure can compute the
    value of any idempotent function (min, max) on any range of the list
    in constant time.

    The structure does not take changes in the input list after
    initialization into account.

    See <https://cp-algorithms.com/data_structures/sparse-table.html>.
    """

    __slots__ = ["sparse_table", "function"]

    def __init__(self, data, function=min):
        """
        Pre-compute the sparse table for range queries.

        Complexity: O(N × log(N)), where N = len(data).

        :param data: input list of objects
        :param function: binary idempotent function to compute
        """
        length = len(data)
        levels = _ilog2(length) + 1

        # sparse_table[depth][i] stores the value of the function
        # on the (i, i + 2**depth) range
        self.sparse_table = [[None] * length for _ in range(levels)]

        if levels > 0:
            self.sparse_table[0] = list(data)

            for depth in range(1, levels):
                for i in range(length - 2**depth + 1):
                    left = self.sparse_table[depth - 1][i]
                    assert left is not None
                    right = self.sparse_table[depth - 1][i + 2 ** (depth - 1)]
                    assert right is not None
                    self.sparse_table[depth][i] = function(left, right)

        self.function = function

    def __call__(self, start: int, stop: int):
        """
        Compute the value of the function on a range.

        Complexity: O(1).

        :param start: first index of the range
        :param stop: index following the last index of the range
        :returns: computed value, or None if the range is empty
        """
        if start >= stop:
            return None

        depth = _ilog2(stop - start)
        return self.function(
            self.sparse_table[depth][start],
            self.sparse_table[depth][stop - 2**depth],
        )



from immutables import Map
from sowing.node import Node
from sowing.zipper import Zipper
from sowing import traversal


def quote_string(data: str) -> str:
    if any(char in "_[](),:;='\t\n" for char in data):
        return "'" + data.replace("'", "''") + "'"

    return data.replace(" ", "_")


def write_props(props: Map) -> str:
    if not props:
        return ""

    return (
        "[&"
        + ",".join(
            f"{quote_string(str(key))}={quote_string(str(value))}"
            for key, value in sorted(props.items())
        )
        + "]"
    )


def write_node(cursor: Zipper[Map | None, Map | None]) -> Zipper[str, None]:
    node = cursor.node
    branch = cursor.data

    if node.edges:
        data = "(" + ",".join(edge.node.data for edge in node.edges) + ")"
    else:
        data = ""

    clade = node.data

    if isinstance(clade, Map):
        if "name" in clade:
            data += quote_string(clade["name"])
            clade = clade.delete("name")

        data += write_props(clade)

    if isinstance(branch, Map) and branch:
        colon_props = []

        for key in ("length", "support", "probability"):
            if key in branch:
                colon_props.append(str(branch[key]))
                branch = branch.delete(key)
            else:
                colon_props.append("")

        while colon_props[-1] == "":
            colon_props.pop()

        other_props = write_props(branch)
        all_props = ":".join(colon_props) + other_props

        if all_props:
            data += ":" + all_props

    return cursor.replace(node=Node(data), data=None)


def write(root: Node[Map | None, Map | None]) -> str:
    """Encode a tree into a Newick string."""
    return traversal.fold(write_node, traversal.depth(root)).data + ";"



from typing import Generic, Hashable, Iterable, Self, TypeVar, overload
from dataclasses import dataclass, replace, field
from collections.abc import Mapping
from .util.dataclasses import repr_default
from .zipper import Zipper


NodeData = TypeVar("NodeData", bound=Hashable)
EdgeData = TypeVar("EdgeData", bound=Hashable)


@repr_default
@dataclass(frozen=True, slots=True)
class Edge(Generic[NodeData, EdgeData]):
    # Node at end of edge
    node: "Node[NodeData, EdgeData]"

    # Arbitrary data attached to this edge
    data: EdgeData | None = None

    def replace(self, **kwargs) -> Self:
        return replace(self, **kwargs)


@repr_default
@dataclass(frozen=True, slots=True)
class Node(Generic[NodeData, EdgeData]):
    # Arbitrary data attached to this node
    data: NodeData | None = None

    # Outgoing edges towards child nodes
    edges: tuple[Edge[NodeData, EdgeData], ...] = ()

    # Cached hash value (to avoid needlessly traversing the whole tree)
    _hash: int = field(init=False, repr=False, compare=False, default=0)

    def __post_init__(self) -> None:
        object.__setattr__(self, "_hash", hash((self.data, self.edges)))

    def __hash__(self) -> int:
        return self._hash

    def replace(self, **kwargs) -> Self:
        return replace(self, **kwargs)

    @overload
    def add(
        self,
        node: Self,
        /,
        *,
        data: EdgeData | None = None,
        index: int = -1,
    ) -> Self:
        """
        Add a new child to this node.

        :param node: new child node
        :param data: optional data to be attached to the linking edge
        :param index: index before which to insert the new child
            (default: insert at the end)
        :returns: updated node
        """
        ...

    @overload
    def add(self, edge: Edge[NodeData, EdgeData], /, *, index: int = -1) -> Self:
        """
        Add an outgoing edge to this node.

        :param edge: new edge to be added
        :param index: index before which to insert the new edge
            (default: insert at the end)
        :returns: updated node
        """
        ...

    def add(
        self,
        node_edge: Self | Edge[NodeData, EdgeData],
        /,
        *,
        data: EdgeData | None = None,
        index: int = -1,
    ) -> Self:
        match node_edge:
            case self.__class__():
                edge = Edge(node=node_edge, data=data)

            case Edge():
                edge = node_edge

            case _:
                raise TypeError(f"cannot add object of type {type(node_edge)}")

        if index == -1:
            index = len(self.edges)

        before = self.edges[:index]
        after = self.edges[index:]
        return self.replace(edges=before + (edge,) + after)

    def extend(
        self,
        items: Iterable[Self | Edge[NodeData, EdgeData]],
    ) -> Self:
        """
        Attach new nodes or edges from an iterable.

        :param items: iterable of nodes or iterable of edges
        :returns: updated node
        """
        for item in items:
            self = self.add(item)

        return self

    def pop(self, index: int = -1) -> Self:
        """
        Remove an outgoing edge from this node.

        :param index: index of the edge to remove
            (default: remove the last one)
        :returns: updated node
        """
        if index == -1:
            index = len(self.edges) - 1

        before = self.edges[:index]
        after = self.edges[index + 1 :]
        return self.replace(edges=before + after)

    def unzip(self) -> Zipper:
        """Make a zipper for this subtree pointing on its root."""
        return Zipper(self)

    def __str__(
        self,
        prefix: str = "",
        chars: dict[str, str] = {
            "root": "┐",
            "branch": "╭",
            "init": "├──",
            "cont": "│  ",
            "init_last": "└──",
            "cont_last": "   ",
            "highlight": "○ ",
        },
        highlight: Self | None = None,
    ) -> str:
        """Create a human-readable representation of this subtree."""
        if self.data is None:
            result = [chars["root"]] if self.edges and not self == highlight else [""]
        elif isinstance(self.data, Mapping):
            result = [str(dict(self.data))]
        else:
            result = [str(self.data)]

        if self == highlight:
            result[0] = chars["highlight"] + result[0]

        init = chars["init"]
        cont = chars["cont"]

        for index, edge in enumerate(self.edges):
            if isinstance(edge.data, Mapping):
                branch = str(dict(edge.data))
            elif edge.data is not None:
                branch = str(edge.data)
            else:
                branch = ""

            if branch:
                result.append(prefix + cont + chars["branch"] + branch)

            if index + 1 == len(self.edges):
                init = chars["init_last"]
                cont = chars["cont_last"]

            subtree = edge.node.__str__(
                prefix=prefix + cont, chars=chars, highlight=highlight
            )
            result.append(prefix + init + subtree)

        return "\n".join(result)



import sys
from immutables import Map
from ..repr import newick
from .. import traversal


def run():
    tree = newick.parse(sys.stdin.read())

    tree = traversal.compress(
        lambda zipper: (
            isinstance(zipper.node.data, Map) and zipper.node.data["name"] == "a"
        ),
        tree,
    )

    print(newick.write(tree))



from ..node import Node, Edge
from .. import traversal
from itertools import product, combinations_with_replacement
from collections import Counter
from typing import Iterable


def is_binary(root: Node) -> bool:
    return all(len(cursor.node.edges) in (0, 2) for cursor in traversal.depth(root))


def binarize_at(root: Node, default: Node = Node()) -> Iterable[Node]:
    """
    Generate all possible binarizations of a node.

    This only binarizes the first level below :param:`root`.
    Use :func:`binarize` to recursively binarize a complete tree.

    Unlabeled, labeled and multiply-labeled trees are supported. Outcomes that
    are isomorphic up to children reorderings are only generated once.

    :param root: original node
    :param default: value used for each new node (default: empty node)
    :returns: iterates through all possible binarizations
    """
    if len(root.edges) <= 2:
        yield root
        return

    head = root.replace(edges=())
    edges = Counter(root.edges)

    # Distribute children multiplicities between left and right subtrees
    for right_muls in product(*(range(count + 1) for count in edges.values())):
        left_muls = tuple(value - mul for mul, value in zip(right_muls, edges.values()))

        # Distribute lexicographically left to right
        if sum(right_muls) == 0 or left_muls < right_muls:
            continue

        left_fan = default
        right_fan = default

        for edge, left_mul, right_mul in zip(edges.keys(), left_muls, right_muls):
            left_fan = left_fan.extend((edge,) * left_mul)
            right_fan = right_fan.extend((edge,) * right_mul)

        left_bin = binarize_at(left_fan, default)
        right_bin = binarize_at(right_fan, default)

        if left_muls == right_muls:
            options = combinations_with_replacement(left_bin, r=2)
        else:
            options = product(left_bin, right_bin)

        for left, right in options:
            yield (
                head.add(left if len(left.edges) > 1 else left.edges[0]).add(
                    right if len(right.edges) > 1 else right.edges[0]
                )
            )


def binarize(root: Node, default: Node = Node()) -> Iterable[Node]:
    """
    Generate all possible binarizations of a tree.

    In the output trees, all internal nodes have at most two children.
    Nodes with three children and more are replaced with all possible
    binary sequences of nodes leading to their children.

    Data on nodes and edges is conserved. Additional nodes and edges
    bear no additional data. Outcomes that are isomorphic up to children
    reorderings are only generated once.

    :param root: original tree
    :param default: value used for each new node (default: empty node)
    :returns: iterates through all possible binarizations
    """
    head = root.replace(edges=())
    children = [edge.node for edge in root.edges]
    edge_data = [edge.data for edge in root.edges]

    for bin_children in product(*(binarize(child, default) for child in children)):
        yield from binarize_at(
            root=head.extend(
                (Edge(node, data) for node, data in zip(bin_children, edge_data))
            ),
            default=default,
        )



from typing import Generic, Hashable, Self, TypeVar, TYPE_CHECKING
from collections.abc import Iterable
from dataclasses import dataclass, replace
from inspect import signature
from .util.dataclasses import repr_default


if TYPE_CHECKING:
    from .node import Node


NodeData = TypeVar("NodeData", bound=Hashable)
EdgeData = TypeVar("EdgeData", bound=Hashable)
OutNodeData = TypeVar("OutNodeData", bound=Hashable)
OutEdgeData = TypeVar("OutEdgeData", bound=Hashable)


@repr_default
@dataclass(frozen=True, slots=True)
class Zipper(Generic[NodeData, EdgeData]):
    # Currently pointed node
    node: "Node[NodeData, EdgeData] | None" = None

    # Data attached to the incoming edge
    data: EdgeData | None = None

    # Child index into the parent node
    index: int = -1

    # Current depth level
    depth: int = 0

    # Parent pointer, or None if at root
    parent: "Zipper[NodeData, EdgeData] | None" = None

    def replace(self, **kwargs) -> Self:
        return replace(self, **kwargs)

    def is_empty(self) -> bool:
        """Test whether there is a pointed node."""
        return self.node is None

    def is_leaf(self) -> bool:
        """Test whether the pointed node is a leaf node."""
        return self.node is None or self.node.edges == ()

    def down(self, index: int = 0) -> "Zipper[NodeData, EdgeData]":
        """
        Move to a child of the pointed node.

        :param index: index of the child to move to;
            negative indices are supported (default: first child)
        :returns: updated zipper
        """
        if self.node is None or index >= len(self.node.edges):
            raise IndexError("child index out of range")

        edges = self.node.edges
        index %= len(edges)
        return self.replace(
            node=edges[index].node,
            data=edges[index].data,
            parent=self.replace(node=self.node.pop(index)),
            index=index,
            depth=self.depth + 1,
        )

    def children(self) -> "Iterable[Zipper[NodeData, EdgeData]]":
        """Iterate through each child of the pointed node."""
        if self.node is None:
            return ()

        return (self.down(index) for index in range(len(self.node.edges)))

    def root(self) -> "Iterable[Zipper[NodeData, EdgeData]]":
        """
        Reconfigure the tree so that the pointed node is the root.

        All nodes, edges and their attached data are preserved. The relative
        node ordering is also preserved. The original tree can be restored
        by re-rooting on the old root node.
        """
        if self.is_root():
            return self

        parent = self.parent.root().node
        parent = parent.replace(
            edges=parent.edges[self.index :] + parent.edges[: self.index]
        )

        return self.replace(
            node=self.node.add(parent, data=self.data),
            data=None,
            index=-1,
            depth=0,
            parent=None,
        )

    def is_root(self) -> bool:
        """Test whether the pointed node is a root node."""
        return self.parent is None

    def up(self) -> "Zipper[NodeData, EdgeData]":
        """Move to the parent of the pointed node."""
        if self.parent is None:
            raise IndexError("cannot go up")

        if self.node is None:
            return self.parent

        if self.parent.node is None:
            raise ValueError("cannot attach to empty parent zipper")

        return self.parent.replace(
            node=self.parent.node.add(self.node, data=self.data, index=self.index),
        )

    def is_last_sibling(self, direction: int = 1) -> bool:
        """
        Test whether the pointed node is the last sibling in a direction.

        :param direction: positive number to test in left to right order,
            negative number to test in right to left order
        """
        if self.parent is None or self.parent.node is None:
            return True

        if direction == 0:
            return False

        if direction < 0:
            return self.index == 0

        return self.index == len(self.parent.node.edges)

    def sibling(self, offset: int = 1) -> "Zipper[NodeData, EdgeData]":
        """
        Move to a sibling of the pointed node.

        :param offset: sibling offset relative to the current node;
            zero for self, positive for right, negative for left, wrapping
            around the child list if needed (default: sibling to the right)
        :returns: updated zipper
        """
        if self.parent is None or self.parent.node is None or offset == 0:
            return self

        if self.is_empty():
            offset -= 1

        index = self.index + offset
        index %= len(self.parent.node.edges) + 1
        return self.up().down(index)

    def _preorder(self, flip: bool) -> "Zipper[NodeData, EdgeData]":
        child = -1 if flip else 0
        sibling = -1 if flip else 1

        if not self.is_leaf():
            return self.down(child)

        while self.is_last_sibling(sibling):
            if self.is_root():
                return self

            self = self.up()

        return self.sibling(sibling)

    def _postorder(self, flip: bool) -> "Zipper[NodeData, EdgeData]":
        child = -1 if flip else 0
        sibling = -1 if flip else 1

        if self.is_root():
            while not self.is_leaf():
                self = self.down(child)

            return self

        if self.is_last_sibling(sibling):
            return self.up()

        self = self.sibling(sibling)

        while not self.is_leaf():
            self = self.down(child)

        return self

    def next(self, preorder: bool = False) -> "Zipper[NodeData, EdgeData]":
        """
        Move to the next node in preorder or postorder.

        :param preorder: pass True to move in preorder (default is postorder)
        :returns: updated zipper
        """
        if preorder:
            return self._preorder(flip=False)
        else:
            return self._postorder(flip=False)

    def prev(self, preorder: bool = False) -> "Zipper[NodeData, EdgeData]":
        """
        Move to the previous node in preorder or postorder.

        :param preorder: pass True to move in preorder (default is postorder)
        :returns: updated zipper
        """
        if preorder:
            return self._postorder(flip=True)
        else:
            return self._preorder(flip=True)

    def zip(self) -> "Node[NodeData, EdgeData] | None":
        """Zip up to the root and return it."""
        bubble = self

        while not bubble.is_root():
            bubble = bubble.up()

        return bubble.node

    def __str__(
        self,
        prefix: str = "",
        chars: dict[str, str] | None = None,
    ) -> str:
        tree = self.zip()

        if chars is None:
            chars = signature(tree.__str__).parameters["chars"].default

        return self.zip().__str__(prefix=prefix, chars=chars, highlight=self.node)



from typing import TypeVar, Generic, Iterable, Self, Generator
from itertools import chain, combinations
from copy import deepcopy


Item = TypeVar("Item")


def powerset(iterable: Iterable[Item]) -> Generator[tuple[Item], None, None]:
    """Generate all subsets of a given sequence."""
    seq = list(iterable)
    return chain.from_iterable(
        combinations(seq, r=size) for size in range(len(seq) + 1)
    )


def groupings(
    iterable: Iterable[Item], size: int = 0
) -> Generator[tuple[tuple[Item]], None, None]:
    """
    Generate all ways to partition a sequence of elements.

    :param iterable: sequence of elements to partition
    :param size: number of subsets in the generated partitions,
        or 0 to generate partitions of any size
    :yields: possible partitions, starting with the largest ones
    """
    seq = tuple(iterable)

    if not seq:
        if size <= 0:
            yield ()

        return

    if size == 1:
        yield (seq,)
        return

    for subset in powerset(seq[1:]):
        subset = (seq[0],) + subset

        for rest in groupings(
            iterable=(item for item in seq if item not in subset),
            size=size - 1,
        ):
            yield (subset, *rest)


class Partition(Generic[Item]):
    """Partition structure implementing the union-find strategy."""

    def __init__(self, items: Iterable[Item] = ()):
        """Create a partition in which each item is in its own set."""
        self._keys = {item: None for item in items}
        self._parent = {item: item for item in self._keys}
        self._rank = {item: 0 for item in self._parent}
        self._count = len(self._parent)

    def copy(self) -> Self:
        """Return a copy of the partition."""
        return deepcopy(self)

    def find(self, item: Item) -> Item:
        """
        Find the group to which an item belongs.

        :returns: a representing item for the group
        """
        root = item

        while self._parent[root] != root:
            root = self._parent[root]

        while item != root:
            item, self._parent[item] = self._parent[item], root

        return root

    def union(self, *items: Item) -> bool:
        """
        Merge items into the same group.

        :param item: any number of items to merge
        :returns: True if and only if at least one group was merged
        """
        merged = False

        if len(items) <= 1:
            return False

        for item2 in items[1:]:
            root1 = self.find(items[0])
            root2 = self.find(item2)

            if root1 == root2:
                continue

            if self._rank[root1] == self._rank[root2]:
                self._parent[root2] = root1
                self._rank[root1] += 1
                del self._keys[root2]

            elif self._rank[root1] > self._rank[root2]:
                self._parent[root2] = root1
                del self._keys[root2]

            else:
                self._parent[root1] = root2
                del self._keys[root1]

            self._count -= 1
            merged = True

        return merged

    def merge(self, size: int = 0) -> Generator[Self, None, None]:
        """
        Generate all possible ways to merge groups of this partition.

        The number of results for a complete partition of `n` elements
        is `B_n`, the n-th Bell number.

        :param size: number of groups in the generated partitions
        :yields: each possible merging
        """
        for grouping in groupings(list(self.keys()), size=size):
            merged = self.copy()

            for group in grouping:
                merged.union(*group)

            yield merged

    def keys(self) -> Iterable[Item]:
        """List the group representants of this partition."""
        return self._keys.keys()

    def values(self) -> Iterable[list[Item]]:
        """List the groups of this partition."""
        result = {key: [] for key in self._keys}

        for item in self._parent:
            result[self.find(item)].append(item)

        return result.values()

    def items(self) -> Iterable[tuple[Item, list[Item]]]:
        return zip(self.keys(), self.values())

    def __repr__(self) -> str:
        items = ", ".join(f"{key!r}: {value!r}" for key, value in self.items())
        return f"{self.__class__.__qualname__}({{{items}}})"

    def __len__(self) -> int:
        """Get the number of groups in this partition."""
        return self._count

    def __eq__(self, other: Self) -> bool:
        return dict(self.items()) == dict(other.items())



from typing import cast, Any, Callable, Generator, Hashable, TypeVar
from functools import partial
from .node import Node
from .zipper import Zipper


T = TypeVar("T")
U = TypeVar("U")

NodeData = TypeVar("NodeData", bound=Hashable)
EdgeData = TypeVar("EdgeData", bound=Hashable)
OutNodeData = TypeVar("OutNodeData", bound=Hashable)
OutEdgeData = TypeVar("OutEdgeData", bound=Hashable)


Traversal = Generator[
    Zipper[NodeData, EdgeData], Zipper[OutNodeData, OutEdgeData], None
]


def _default(value: T) -> Generator[T, U | None, U | T]:
    result = yield value
    return result if result is not None else value


def depth(
    node: Node[NodeData, EdgeData] | None,
    preorder: bool = False,
    reverse: bool = False,
) -> Traversal[NodeData, EdgeData, OutNodeData, OutEdgeData]:
    """
    Traverse a tree in depth-first order.

    :param node: root node to start from
    :param preorder: pass True to visit parents before children (preorder),
        defaults to children before parents (postorder)
    :param reverse: pass True to reverse the order
    :returns: generator that yields nodes in the specified order
    """
    if node is None:
        return

    cursor = node.unzip()
    advance = partial(
        Zipper.prev if reverse else Zipper.next,
        preorder=preorder,
    )
    root_start = not preorder == reverse

    if not root_start:
        cursor = advance(cursor)

    while True:
        cursor = yield from _default(cursor)

        if not root_start and cursor.is_root():
            return

        cursor = advance(cursor)

        if root_start and cursor.is_root():
            return


def euler(
    node: Node[NodeData, EdgeData] | None,
    reverse: bool = False,
) -> Traversal[NodeData, EdgeData, OutNodeData, OutEdgeData]:
    """
    Traverse a tree along an eulerian tour.

    :param node: root node to start from
    :param reverse: pass True to reverse the order
    :returns: generator that yields nodes in the specified order
    """
    if node is None:
        return

    child = -1 if reverse else 0
    sibling = -1 if reverse else 1
    cursor = node.unzip()

    while True:
        cursor = yield from _default(cursor)

        if not cursor.is_leaf():
            cursor = cursor.down(child)
        else:
            while cursor.is_last_sibling(sibling) and not cursor.is_root():
                cursor = cursor.up()
                cursor = yield from _default(cursor)

            if cursor.is_root():
                return
            else:
                pos = cursor.index
                cursor = cursor.up()
                cursor = yield from _default(cursor)
                cursor = cursor.down(pos + sibling)


def leaves(
    node: Node[NodeData, EdgeData] | None,
    reverse: bool = False,
) -> Traversal[NodeData, EdgeData, OutNodeData, OutEdgeData]:
    """
    Traverse the leaves of a tree.

    :param node: root node to start from
    :param reverse: pass True to reverse the order
    :returns: generator that yields leaves in the specified order
    """
    if node is None:
        return

    advance = Zipper.prev if reverse else Zipper.next
    cursor = advance(node.unzip())

    while True:
        if cursor.is_leaf():
            cursor = yield from _default(cursor)

        if cursor.is_root():
            return

        cursor = advance(cursor)


def fold(
    func: Callable[[Zipper[NodeData, EdgeData]], Zipper[OutNodeData, OutEdgeData]],
    traversal: Traversal[NodeData, EdgeData, OutNodeData, OutEdgeData],
) -> Node[OutNodeData, OutEdgeData] | None:
    """
    Transform a tree along a given traversal.

    For each node along the traversal, the folding callback is invoked with
    a cursor pointing on this node. The callback can transform this cursor in
    any way (including changing the tree structure) and returns the updated
    cursor, which is used as a starting point to continue the traversal.

    :param func: callback receiving zipper values along the traversal
        and returning an updated zipper
    :param traversal: tree traversal generator
    :returns: transformed tree
    """
    try:
        in_cursor = next(traversal)
    except StopIteration:
        return None

    cursor = cast(Zipper[Any, Any], in_cursor)

    try:
        while True:
            cursor = func(cursor)
            cursor = traversal.send(cursor)
    except StopIteration:
        out_cursor = cast(Zipper[OutNodeData, OutEdgeData], cursor)
        return out_cursor.zip()


def map(
    func: Callable[[NodeData, EdgeData, int, int], tuple[OutNodeData, OutEdgeData]],
    traversal: Traversal,
) -> Node[OutNodeData, OutEdgeData] | None:
    """
    Map values attached to nodes and edges along a given tree traversal.

    For each node along the traversal, the mapping callback is invoked with

    (1) the data object attached to the node,
    (2) the data object attached to its parent edge,
    (3) the node’s sibling index relative to its parent, and
    (4) its depth in the tree.

    The mapping callback returns a tuple of two values, the updated data object
    for the node and for the parent edge.

    :param func: mapping callback
    :param traversal: tree traversal
    :returns: transformed tree
    """

    def wrapper(zipper: Zipper[NodeData, EdgeData]) -> Zipper[OutNodeData, OutEdgeData]:
        node, edge = func(zipper.node.data, zipper.data, zipper.index, zipper.depth)
        return zipper.replace(node=zipper.node.replace(data=node), data=edge)

    return fold(wrapper, traversal)


def compress(
    func: Callable[[Zipper[NodeData, EdgeData]], bool],
    node: Node[NodeData, EdgeData] | None,
) -> Node[NodeData, EdgeData] | None:
    """
    Compress non-root nodes matching a filter, attaching their children to their parent.

    :param func: filter to select nodes to compress
    :param node: tree to filter
    :returns: transformed tree
    """

    def wrapper(zipper: Zipper[NodeData, EdgeData]) -> Zipper[NodeData, EdgeData]:
        result = zipper.node.replace(edges=())

        for child in zipper.children():
            if func(child):
                result = result.extend(child.node.edges)
            else:
                result = result.add(child.node, data=child.data)

        return zipper.replace(node=result)

    return fold(wrapper, depth(node, preorder=False))



from dataclasses import dataclass
from typing import Set, Generator
from itertools import product
from .. import traversal
from ..node import Node
from ..zipper import Zipper
from ..util.partition import Partition


@dataclass(frozen=True, slots=True)
class Triple:
    ingroup: tuple[Node, Node]
    outgroup: Node

    def is_in(self, nodes) -> bool:
        return (set(self.ingroup) | set((self.outgroup,))) <= set(nodes)


@dataclass(frozen=True, slots=True)
class Fan:
    group: tuple[Node, ...]

    def is_in(self, nodes) -> bool:
        return set(self.group) <= set(nodes)


def breakup(root: Node) -> tuple[list[Node], list[Triple], list[Fan]]:
    """
    Break up a phylogenetic tree into triples and fans that encode its topology.

    This implements the BreakUp algorithm from [Ng and Wormald, 1996].

    The output representation uniquely determines the input tree, disregarding
    unary nodes, repeated leaves, child order, data in internal nodes, and
    data on edges.

    The input tree can be reconstructed using the :func:`build` function.

    :param root: input tree to be broken up
    :returns: a tuple containing the list of leaves in the tree, a list of
        extracted triples, and a list of extracted fans
    """
    triples = []
    fans = []

    def extract_parts(cursor: Zipper) -> Zipper:
        children = tuple(edge.node for edge in cursor.node.edges)

        if cursor.is_leaf() or not all(
            cursor.down(i).is_leaf() for i in range(len(children))
        ):
            return cursor

        if len(children) >= 3:
            # Break up fan
            fans.append(Fan(children))
            children = children[:2]

        # Break up triple
        base = cursor

        while base.sibling() == base and not base.is_root():
            base = base.up()

        if base.is_root():
            return cursor

        outgroup = next(traversal.leaves(base.sibling().node)).node
        triples.append(Triple(children, outgroup))
        return base.replace(node=children[0])

    traversal.fold(extract_parts, traversal.depth(root))
    leaves = [cursor.node for cursor in traversal.leaves(root)]
    return leaves, triples, fans


def _build_merge_fans(partition, fans):
    """Ensure fans are either all in the same group or all in different groups."""
    merged = True

    while merged:
        merged = False

        for fan in fans:
            leaves = list(fan.group)
            roots = {partition.find(leaf) for leaf in leaves}

            if len(roots) < len(leaves):
                merged = merged or partition.union(*leaves)


def build(
    leaves: list[Node],
    triples: list[Triple] = [],
    fans: list[Fan] = [],
    arity: int = 0,
) -> Generator[Node, None, None]:
    """
    Enumerate phylogenetic trees satisfying the topology constraints given by
    a set of triples and fans.

    The first returned tree is the smallest tree (in number of nodes) compatible
    with all the triples and fans given as input, if such a tree exists.

    This implements the AllTrees algorithm from [Ng and Wormald, 1996].

    :param leaves: set of leaves
    :param triples: set of triples
    :param fans: set of fans
    :param arity: arity of the generated trees, or 0 to generate trees of arbitrary arity
    :yields: possible trees, if any
    """
    if not leaves:
        return

    if len(leaves) == 1:
        yield leaves[0]
        return

    if len(leaves) == 2:
        left, right = leaves
        yield Node().add(left).add(right)
        return

    partition = Partition(leaves)

    # Merge groups for triples
    for triple in triples:
        partition.union(*triple.ingroup)

    _build_merge_fans(partition, fans)

    # Try all possible mergings of the partition that fit the requested arity
    for subpartition in partition.merge(size=arity):
        _build_merge_fans(subpartition, fans)

        if len(subpartition) <= 1:
            return

        # Recursively build subtrees for each group
        for descendants in product(
            *(
                build(
                    group,
                    [triple for triple in triples if triple.is_in(group)],
                    [fan for fan in fans if fan.is_in(group)],
                    arity=arity,
                )
                for group in subpartition.values()
            )
        ):
            yield Node().extend(descendants)


def supertree(*trees: Node, arity: int = 0) -> Generator[Node, None, None]:
    """
    Build a supertree from a set of phylogenetic trees.

    The first returned tree is the smallest tree compatible with every tree of
    the input, if such a tree exists.

    :param tree: any number of tree to build a supertree from
    :param arity: arity of the generated supertrees, or 0 to generate
        supertrees of arbitrary arity
    :yields: possible supertrees, if any
    """
    # Use dictionaries as sets to merge parts while preserving ordering
    all_leaves = {}
    all_triples = {}
    all_fans = {}

    for tree in trees:
        leaves, triples, fans = breakup(tree)
        all_leaves.update(dict.fromkeys(leaves))
        all_triples.update(dict.fromkeys(triples))
        all_fans.update(dict.fromkeys(fans))

    yield from build(
        all_leaves.keys(), all_triples.keys(), all_fans.keys(), arity=arity
    )


def display(root: Node, leaves: Set[Node]) -> Node | None:
    """
    Extract the smallest minor containing the given leaves of a tree.

    :param root: input tree to be filtered
    :param leaves: set of leaves to keep
    :returns: filtered tree
    """

    def filter_tree(cursor: Zipper) -> Zipper:
        node = cursor.node

        if cursor.is_leaf() and node not in leaves:
            return cursor.replace(node=None)

        if len(node.edges) == 1:
            return cursor.replace(node=node.edges[0].node)

        return cursor

    return traversal.fold(filter_tree, traversal.depth(root))



from typing import Any, Callable, Hashable, TypeVar
from json import JSONEncoder
from ..node import Node, Edge


NodeData = TypeVar("NodeData", bound=Hashable)
EdgeData = TypeVar("EdgeData", bound=Hashable)


class TreeEncoder(JSONEncoder):
    def default(self, value):
        match value:
            case Node(data, edges):
                return {"edges": edges, "data": data}

            case Edge(node, data):
                return {"node": node, "data": data}

        return super().default(value)


def passthrough(value):
    return value


def tree_decoder(
    value: dict,
    node_decoder: Callable[[Any], NodeData] = passthrough,
    edge_decoder: Callable[[Any], EdgeData] = passthrough,
) -> Node[NodeData, EdgeData] | Edge[NodeData, EdgeData]:
    if not isinstance(value, dict):
        raise TypeError("value must be a dict, got {type(value)}")

    if "edges" in value:
        if not isinstance(value["edges"], list):
            raise TypeError("node edges must be a list, got {type(value['edges'])}")

        return Node(
            edges=tuple(
                tree_decoder(item, node_decoder, edge_decoder)
                for item in value["edges"]
            ),
            data=node_decoder(value.get("data")),
        )

    if "node" in value:
        return Edge(
            node=tree_decoder(value["node"], node_decoder, edge_decoder),
            data=edge_decoder(value.get("data")),
        )

    raise TypeError(
        "value is neither a node with an 'edges' key nor"
        f" an edge with a 'node' key; got keys {list(value.keys())}"
    )
